#include "ApeDecoder.h"

CApeIO::CApeIO(const wchar_t* filename, bool usemem)
{
	m_usemem=usemem;
	if(m_stream.Open(filename,usemem))
	{
		m_filename=filename;
		m_opened=1;
	}
	else
		m_opened=0;
}

CApeIO::~CApeIO()
{
	if(m_opened)
		m_stream.Close();
}

int CApeIO::Open(const wchar_t* filename)
{
	return m_opened?0:-1;
}

int CApeIO::Close()
{
	return 0;
}

int CApeIO::Read(void * pBuffer, unsigned int nBytesToRead, unsigned int * pBytesRead)
{
	*pBytesRead=m_stream.Read(pBuffer,nBytesToRead);
	return 0;
}

int CApeIO::Seek(int nDistance, unsigned int nMoveMode)
{
	m_stream.Seek(nDistance,nMoveMode);
	return 0;
}

int CApeIO::GetPosition()
{
	return (int)m_stream.Tell();
}

int CApeIO::GetSize()
{
	return (int)m_stream.Length();
}

int CApeIO::GetName(wchar_t * pBuffer)
{
	wcscpy(pBuffer,m_filename.c_str());
	return 0;
}

CApeDecoder::CApeDecoder()
{
	m_io=0;
	m_ape=0;
	m_opened=0;
}

CApeDecoder::CApeDecoder(const wchar_t* filename)
{
	m_error[0]=0;
	m_opened=0;
	m_io=new CApeIO(filename,m_usemem);
	m_ape=CreateIAPEDecompressEx(m_io);
	if(m_ape)
	{
		m_info=(APE_FILE_INFO*)m_ape->GetInfo(APE_INTERNAL_INFO,0,0);
		if(m_info&&m_info->nVersion>=3980&&m_info->spAPEDescriptor)
		{
			m_md5=1;
			memcpy(m_md5check,m_info->spAPEDescriptor->cFileMD5,16);
			md5_init(&m_md5state);
			m_io->Seek(m_info->nJunkHeaderBytes+m_info->spAPEDescriptor->nDescriptorBytes+m_info->spAPEDescriptor->nHeaderBytes+m_info->spAPEDescriptor->nSeekTableBytes,SEEK_SET);
			m_samples=m_info->spAPEDescriptor->nHeaderDataBytes+m_info->spAPEDescriptor->nAPEFrameDataBytes+m_info->spAPEDescriptor->nTerminatingDataBytes;
			m_blocks=0;
			m_opened=1;
		}
		else
		{
			m_md5=0;
			m_samples=m_ape->GetInfo(APE_INFO_TOTAL_BLOCKS,0,0);
			m_blocks=m_ape->GetInfo(APE_INFO_BLOCK_ALIGN,0,0);
			if(m_blocks)
			{
				m_blocks=APESIZE/m_blocks;
				m_opened=1;
			}
		}
	}
}

CApeDecoder::~CApeDecoder()
{
	delete m_io;
	delete m_ape;
}

const wchar_t* CApeDecoder::GetSupportedTypes()
{
	return APETYPES;
}

CDecoder* CApeDecoder::NewDecoder(const wchar_t* filename)
{
	CApeDecoder* d=new CApeDecoder(filename);
	if(d->m_opened)
		return d;
	else
	{
		delete d;
		return 0;
	}
}

__int64 CApeDecoder::GetSampleCount()
{
	return m_samples;
}

const wchar_t* CApeDecoder::GetLastError()
{
	return m_error;
}

long CApeDecoder::Read()
{
	int i;
	if(m_md5)
	{
		i=(m_samples-m_blocks<APESIZE)?m_samples-m_blocks:APESIZE;
		m_io->Read(m_buffer,i,(unsigned int*)&i);
		md5_append(&m_md5state,(unsigned char*)m_buffer,i);
		m_blocks+=i;
		if(!i)
		{
			unsigned char md5[16];
			m_io->Seek(m_info->nJunkHeaderBytes+m_info->spAPEDescriptor->nDescriptorBytes,SEEK_SET);
			i=m_info->spAPEDescriptor->nHeaderBytes+m_info->spAPEDescriptor->nSeekTableBytes;
			if(i&&i<APESIZE)
			{
				m_io->Read(m_buffer,i,(unsigned int*)&i);
				md5_append(&m_md5state,(unsigned char*)m_buffer,i);
			}
			md5_finish(&m_md5state,md5);
			if(memcmp(md5,m_md5check,16))
			{
				wcscpy(m_error,L"MD5 MISMATCH");
				i=-1;
			}
			else
				i=0;
		}
		return i;
	}
	else
	{
		int error;
		if(error=m_ape->GetData(m_buffer,m_blocks,&i))
		{
			switch(error)
			{
			case ERROR_IO_READ:
				wcscpy(m_error,L"I/O READ ERROR");
				break;
			case ERROR_INVALID_INPUT_FILE:
				wcscpy(m_error,L"INVALID INPUT FILE");
				break;
			case ERROR_INSUFFICIENT_MEMORY:
				wcscpy(m_error,L"INSUFFICIENT MEMORY");
				break;
			case ERROR_INVALID_CHECKSUM:
				error=m_ape->GetInfo(APE_DECOMPRESS_CURRENT_MS,0,0)/1000;
				swprintf(m_error,MAXERROR,L"INVALID CHECKSUM @ %dm %02ds",error/60,error%60);
			}	
			return -1;
		}
		else
			return i;
	}
}
