#pragma once

#include "decoders.h"
#include "stream.h"
#include "md5.h"
#include <ape/all.h>
#include <ape/MACLib.h>
#include <ape/IO.h>

#define APESIZE	65536
const wchar_t APETYPES[]=L"ape\0";

struct APE_FILE_INFO
{
    int nVersion;                                   // file version number * 1000 (3.93 = 3930)
    int nCompressionLevel;                          // the compression level
    int nFormatFlags;                               // format flags
    int nTotalFrames;                               // the total number frames (frames are used internally)
    int nBlocksPerFrame;                            // the samples in a frame (frames are used internally)
    int nFinalFrameBlocks;                          // the number of samples in the final frame
    int nChannels;                                  // audio channels
    int nSampleRate;                                // audio samples per second
    int nBitsPerSample;                             // audio bits per sample
    int nBytesPerSample;                            // audio bytes per sample
    int nBlockAlign;                                // audio block align (channels * bytes per sample)
    int nWAVHeaderBytes;                            // header bytes of the original WAV
    int nWAVDataBytes;                              // data bytes of the original WAV
    int nWAVTerminatingBytes;                       // terminating bytes of the original WAV
    int nWAVTotalBytes;                             // total bytes of the original WAV
    int nAPETotalBytes;                             // total bytes of the APE file
    int nTotalBlocks;                               // the total number audio blocks
    int nLengthMS;                                  // the length in milliseconds
    int nAverageBitrate;                            // the kbps (i.e. 637 kpbs)
    int nDecompressedBitrate;                       // the kbps of the decompressed audio (i.e. 1440 kpbs for CD audio)
    int nJunkHeaderBytes;                           // used for ID3v2, etc.
    int nSeekTableElements;                         // the number of elements in the seek table(s)

    CSmartPtr<uint32> spSeekByteTable;              // the seek table (byte)
    CSmartPtr<unsigned char> spSeekBitTable;        // the seek table (bits -- legacy)
    CSmartPtr<unsigned char> spWaveHeaderData;      // the pre-audio header data
    CSmartPtr<APE_DESCRIPTOR> spAPEDescriptor;      // the descriptor (only with newer files)
};

class CApeIO : public CIO
{
public:
	CApeIO(const wchar_t* filename, bool usemem);
	virtual ~CApeIO();

	// open / close
	virtual int Open(const wchar_t * pName);
	virtual int Close();

	// read / write
	virtual int Read(void * pBuffer, unsigned int nBytesToRead, unsigned int * pBytesRead);
	virtual int Write(const void * pBuffer, unsigned int nBytesToWrite, unsigned int * pBytesWritten){return 0;}

	// seek
	virtual int Seek(int nDistance, unsigned int nMoveMode);

	// creation / destruction
	virtual int Create(const wchar_t * pName){return 0;}
	virtual int Delete(){return 0;}

	// other functions
	virtual int SetEOF(){return 0;}

	// attributes
	virtual int GetPosition();
	virtual int GetSize();
	virtual int GetName(wchar_t * pBuffer);

private:
	bool m_usemem;
	bool m_opened;
	CStream m_stream;
	wstring m_filename;
};

class CApeDecoder : public CDecoder
{
public:
	CApeDecoder();
	CApeDecoder(const wchar_t* filename);
	virtual ~CApeDecoder();

	virtual const wchar_t* GetSupportedTypes();
	virtual CDecoder* NewDecoder(const wchar_t* filename);

	virtual __int64 GetSampleCount();
	virtual const wchar_t* GetLastError();
	virtual long Read();

	bool m_opened;

private:
	CApeIO* m_io;
	IAPEDecompress* m_ape;
	APE_FILE_INFO* m_info;
	__int64 m_samples;
	unsigned int m_blocks;
	bool m_md5;
	md5_state_t m_md5state;
	unsigned char m_md5check[16];
	wchar_t m_error[MAXERROR];
	char m_buffer[APESIZE];
};
